#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
"""
@File    :   memory.py
@Time    :   2024-08-04 19:38
@Author  :   坐公交也用券
@Version :   1.0
@Contact :   faith01238@hotmail.com
@Homepage : https://liumou.site
@Desc    :   定时调整虚拟机内存
"""

import logging
import os
import subprocess
import sys
import tomllib
from argparse import ArgumentParser
from datetime import datetime
from os import path
from sys import exit
from  time import sleep
# 创建一个自定义的日志记录器
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# 创建一个自定义的日志处理器，设置其输出格式
formatter = logging.Formatter('%(asctime)s [%(levelname)s] %(filename)s:%(lineno)d | %(funcName)s | %(message)s')
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(formatter)
logger.addHandler(handler)


def list_files(directory):
	"""
	列出指定目录下的所有文件
	"""
	file_list = []
	# 使用glob模块获取指定目录下的所有文件
	for root, dirs, files in os.walk(directory):
		for name in files:
			# print(os.path.join(root, name))
			file_list.append(os.path.join(root, name))
	return file_list


class QmSetMen:
	def __init__(self, config_list: str, debug: bool = False):
		"""

		:param config_list:
		:param debug:
		"""
		self.config_list = config_list
		self.debug = debug
		self.txt_config = {}
		self.txt_config_list = {}
		self.log_file = None
		# 当前时间:分
		self.current_min = 0
		# 当前时间:时
		self.current_hour = 0
		if self.debug:
			logger.debug("开始实例化")
		# 设置实例配置
		self.toml_name = ""
		self.toml_id = 220
		self.toml_mem_size = 0
		# 开启操作
		self.toml_start_hour = 0
		self.toml_on_delayed = 0

	def get_time(self):
		"""
		获取当前时间
		:return:
		"""
		if self.debug:
			logger.debug("正在获取当前时间")
		current = datetime.now()
		self.current_hour = current.hour
		self.current_min = current.minute
		if self.debug:
			logger.debug(f"当前小时: {self.current_hour} - 当前分钟: {self.current_min}")

	def read_list(self, file):
		"""
		读取配置文件
		:param file: 需要读取的主机配置文件
		:return:
		"""
		if self.debug:
			logger.debug(f"正在读取配置文件: {file}")
		try:
			with open(file, "rb") as f:
				self.txt_config_list = tomllib.load(f)
				return
		except Exception as e:
			print(file)
			logger.error(f"读取配置文件失败: {e}")
			exit(1)

	def result_log(self, result: str):
		"""
		输出结果
		:return:
		"""
		if self.debug:
			logger.debug("正在输出结果")
		if self.log_file:
			logger.info(f"日志文件: {self.log_file}")
		else:
			logger.info("日志文件: None")
			return
		if not path.isfile(self.log_file):
			os.system(f"touch {self.log_file}")
		try:
			with open(file=self.log_file, mode='a+', encoding='utf-8') as w:
				w.write(result)
		except Exception as e:
			logger.error(f"写入日志文件失败: {e}")

	def qm_set_mem(self, Id, Size: int):
		"""
		设置虚拟机内存
		:param Size: 内存大小
		:param Id: 虚拟机ID
		:return:
		"""
		size_ = 0
		if self.debug:
			logger.debug(f"正在获取当前配置: {Id}")
		cmd = str("qm config %d | grep memory| awk '{print $2}'" % Id)
		if self.debug:
			print(size_)
			print(cmd)
		res = subprocess.getstatusoutput(cmd)
		if res[0] == 0:
			res_out = res[1].strip().lower()

			try:
				size_ = int(res_out)
			except Exception as e:
				logger.error(f"获取内存大小失败: {e}")
				return
			if self.debug:
				logger.debug(f"当前大小: {res_out}")
				logger.debug(f"目标大小: {Size}")
			if int(size_) == int(Size):
				logger.info(f"当前内存大小: {size_} - 目标内存大小: {Size} - 不需要修改")
				return
			else:
				if os.system(f"qm stop {Id}") != 0:
					logger.error(f"停止虚拟机失败: {Id}")
					return
				cmd = str(f'qm set  {Id} --memory {Size}')
				if os.system(cmd) != 0:
					logger.error(f"设置内存大小失败: {Id}")
					return
			# 判断是否已经开机
			cmd = str(f"qm status {Id} " + " | awk '{print $2}'")
			if self.debug:
				logger.debug(f"正在获取状态: {Id}")
			res = subprocess.getstatusoutput(cmd)
			if res[0] == 0:
				if res[1].strip().lower() == "stopped":
					if self.debug:
						logger.debug(f"正在开机: {Id}")
					# 开机
					ops = "开机"
					cmd_op = "start"
					if self.debug:
						logger.debug(f"正在{ops}: {Id}")
					if os.system(f"qm {cmd_op} {Id}") == 0:
						logger.info(f"{ops}成功: {Id}")
						self.result_log(f"{datetime.now()} - {Id}  - {ops}成功\n")
					else:
						logger.error(f"{ops}失败: {Id}")
						self.result_log(f"{datetime.now()} - {Id}  - {ops}失败\n")
				else:
					logger.info(f"当前状态: {res[1].strip().lower()} - 不需要执行开机指令")
		else:
			logger.error(f"获取状态失败: {Id}")
			print(res[1])
		sleep(20)


	def check_time(self, start_hour, start_min, end_hour, end_min):
		"""
		判断时间是否在配置文件设置的时间范围内
		:param start_hour:
		:param start_min:
		:param end_hour:
		:param end_min:
		:return:
		"""
		# 格式化成时间对象
		start_time = datetime.strptime(f"{start_hour}:{start_min}", "%H:%M")
		end_time = datetime.strptime(f"{end_hour}:{end_min}", "%H:%M")
		# 获取当前时间对象，仅获取小时和分钟
		start_time = start_time.strftime("%H:%M")
		end_time = end_time.strftime("%H:%M")
		now_time = datetime.now().strftime("%H:%M")
		if self.debug:
			logger.debug(f"当前时间: {now_time}")
		# 如果当前时间在配置文件设置的时间范围内，则返回True，否则返回False
		if start_time <= now_time <= end_time:
			if self.debug:
				logger.debug(f"当前时间在配置文件设置的时间范围内: {start_time} - {end_time}")
			return True
		else:
			if self.debug:
				logger.debug(f"当前时间不在配置文件设置的时间范围内: {start_time} - {end_time}")
			return False

	def list_config(self):
		"""
		读取列表主机配置文件
		:return:
		"""
		# 判断文件夹是否存在且是否有文件
		if path.isdir(self.config_list):
			fl = list_files(self.config_list)
			if fl:
				if self.debug:
					# 打印文件数量
					logger.debug(f"当前配置文件数量: {len(fl)}")
				for file in fl:
					self.read_list(file=file)
					# 获取需要操作的主机ID列表
					data_list = self.txt_config_list["list"]
					host_list = data_list["list"]
					hour_start_ = data_list["start_hour"]
					min_start_ = data_list["start_minute"]
					hour_end_ = data_list["end_hour"]
					min_end = data_list["end_minute"]
					size_ = data_list["size"]
					enable_ = data_list["enabled"]
					if self.check_time(hour_start_, min_start_, hour_end_, min_end) and enable_:
						for i in host_list:
							self.qm_set_mem(i, size_)

			else:
				logger.warning(f"配置文件目录为空: {self.config_list}")
		else:
			if self.debug:
				logger.debug(f"配置文件目录不存在: {self.config_list}")

	def start(self):
		while True:
			self.list_config()


def install():
	"""
	安装
	:return:
	"""
	memory_bin = path.join("/usr/bin/com.liumou.pve.qm.memory")
	memory_url = "https://gitee.com/liumou_site/pvet/raw/master/model/qm/memory.py"
	logger.info("正在进行安装/更新...")
	url_list = 'https://gitee.com/liumou_site/pvet/raw/master/model/qm/memory.toml'
	os.system("mkdir -p /opt/apps/com.liumou.pve.qm.memory/")
	demo_list = "/opt/apps/com.liumou.pve.qm.memory/demo_list.toml"
	url_dict = {demo_list: url_list, memory_bin: memory_url}
	for file in url_dict:
		if path.exists(file):
			logger.info("模板配置文件已存在")
			os.system(f"rm -f {file}")
		res = subprocess.getstatusoutput(f'wget -O {file} {url_dict[file]}')
		if res[0] == 0:
			logger.info(f"[ {file} ]配置文件下载成功")
		else:
			logger.info(f"[ {file} ]配置文件下载失败")
			print(res[1])
			exit(1)
	os.system(f"chmod 0755 {memory_bin}")
	logger.info("安装/更新成功")
	service()
	exit(0)


def service():
	"""

	:return:
	"""
	if not path.isfile("/usr/bin/com.liumou.pve.qm.memory"):
		logger.error(f"执行文件不存在: /usr/bin/com.liumou.pve.qm.memory")
		exit(1)
	txt = f"""[Unit]
Description=This is a PVE timing tool
After=network.target

[Service]
Type=simple
User=root
ExecStart=/usr/bin/com.liumou.pve.qm.memory
StandardOutput=journal
KillMode=control-group
Restart=on-failure
RestartPreventExitStatus=255
StandardError=inherit
SyslogIdentifier=com.liumou.pve.qm.memory

[Install]
WantedBy=multi-user.target
Alias=com.liumou.pve.qm.memory.service
"""
	try:
		with open(file="/etc/systemd/system/com.liumou.pve.qm.memory.service", mode='w', encoding='utf-8') as w:
			w.write(txt)
	except Exception as e:
		logger.error(f"写入服务文件失败: {e}")
		exit(1)
	if os.system("systemctl daemon-reload") == 0:
		os.system("systemctl enable com.liumou.pve.qm.memory.service")
		if os.system("systemctl restart com.liumou.pve.qm.memory.service") == 0:
			logger.info("服务启动成功")
			exit(0)
		else:
			logger.error("服务启动失败...")
			exit(3)
	else:
		logger.error("服务启动失败")
		exit(2)


if __name__ == '__main__':
	conf_default_list = "/opt/apps/com.liumou.pve.qm.memory/config_list/"
	os.system(f"mkdir -p /opt/apps/com.liumou.pve.qm.memory/log/ {conf_default_list}")
	arg = ArgumentParser(description='当前脚本版本:1.0', prog="PVE虚拟机/容器内存定时调整")
	h = f"设置列表配置文件路径,默认: {conf_default_list}"
	arg.add_argument('-cl', '--conf_list', type=str, help=h, default=conf_default_list, required=False)
	arg.add_argument('-f', action='store_true', help='强制更新配置文件')
	arg.add_argument('-debug', action='store_true', help='开启调试模式')
	arg.add_argument('-install', action='store_true', help='安装/更新服务')
	args = arg.parse_args()
	args_conf_list = args.conf_list
	args_install = args.install
	args_debug = args.debug
	args_f = args.f
	if args_debug:
		logger.info("开启了调试模式")
	if args_install:
		logger.info("开启了安装/更新功能")
		install()
	c = QmSetMen(debug=args_debug, config_list=args_conf_list)
	c.start()
